@@ -1,6 +1,6 @@
 /*
  * Intel(R) Enclosure LED Utilities
- * Copyright (C) 2009-2013 Intel Corporation.
+ * Copyright (C) 2009-2018 Intel Corporation.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms and conditions of the GNU General Public License,
@@ -30,7 +30,7 @@
  * - FAIL        - a block device has failed or is missing. Failure LED is
  *                 active and the behavior is depended on implementation
  *                 of enclosure management processor.
- * - REBUILD(_P) - this means a RAID device is recovering or rebuilding
+ * - REBUILD     - this means a RAID device is recovering or rebuilding
  *                 its data. Depending on implementation of enclosure
  *                 management processor appropriate LED is blinking or solid.
  * - ICA         - In a Critical Array, this means a RAID device is degraded and
@@ -52,6 +52,10 @@
  *                 is set, when a RAID device disappears, too. Oneshot means
  *                 as soon application applies the state it will change
  *                 to UNKNOWN.
+ * - ADDED         this state means that device previously known to ledmon is
+ *                 restored. This state will be changed to ONESHOT_NORMAL.
+ * - REMOVED       this state means that device was removed from system. It
+ *                 will be changed to ADDED after restoring device to system.
  */
 enum ibpi_pattern {
 	IBPI_PATTERN_UNKNOWN = 0,
@@ -61,12 +65,13 @@ enum ibpi_pattern {
 	IBPI_PATTERN_DEGRADED,
 	IBPI_PATTERN_HOTSPARE,
 	IBPI_PATTERN_REBUILD,
-	IBPI_PATTERN_REBUILD_P,
 	IBPI_PATTERN_FAILED_ARRAY,
 	IBPI_PATTERN_PFA,
 	IBPI_PATTERN_FAILED_DRIVE,
 	IBPI_PATTERN_LOCATE,
 	IBPI_PATTERN_LOCATE_OFF,
+	IBPI_PATTERN_ADDED,
+	IBPI_PATTERN_REMOVED,
 	/* Below are SES-2 codes. Note that by default most IBPI messages are
 	 * translated into SES when needed but SES codes can be added also. */
 	SES_REQ_ABORT,
@@ -86,7 +91,8 @@ enum ibpi_pattern {
 	SES_REQ_EN_BB,
 	SES_REQ_EN_BA,
 	SES_REQ_DEV_OFF,
-	SES_REQ_FAULT
+	SES_REQ_FAULT,
+	SES_REQ_PRDFAIL,
 };
 
 extern const char *ibpi_str[];
