@@ -1,6 +1,6 @@
 /*
  * Intel(R) Enclosure LED Utilities
- * Copyright (C) 2009-2013 Intel Corporation.
+ * Copyright (C) 2009-2018 Intel Corporation.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms and conditions of the GNU General Public License,
@@ -17,52 +17,46 @@
  *
  */
 
-#include <config.h>
-#include <unistd.h>
-#include <stdlib.h>
-#include <syslog.h>
-#include <fcntl.h>
-#include <limits.h>
-#include <time.h>
-#include <signal.h>
-#include <stdio.h>
 #include <errno.h>
-#include <string.h>
+#include <fcntl.h>
 #include <getopt.h>
+#include <limits.h>
+#include <signal.h>
 #include <stdint.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <sys/param.h>
+#include <sys/select.h>
+#include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/wait.h>
-#include <sys/stat.h>
-#include <sys/select.h>
+#include <syslog.h>
+#include <time.h>
+#include <unistd.h>
 
 #if _HAVE_DMALLOC_H
 #include <dmalloc.h>
 #endif
 
-#include "status.h"
-#include "ibpi.h"
-#include "pidfile.h"
-#include "list.h"
-#include "utils.h"
-#include "sysfs.h"
-#include "block.h"
-#include "slave.h"
-#include "raid.h"
-#include "cntrl.h"
-#include "version.h"
-#include "scsi.h"
 #include "ahci.h"
+#include "block.h"
+#include "cntrl.h"
+#include "config.h"
+#include "config_file.h"
+#include "ibpi.h"
+#include "list.h"
+#include "pidfile.h"
+#include "raid.h"
+#include "scsi.h"
+#include "slave.h"
 #include "smp.h"
-
-/**
- * Default suspend time between sysfs scan operations, given in seconds.
- */
-#define DEFAULT_SLEEP_INTERVAL    10
-
-/**
- * Minimum suspend time between sysfs scan operations, given in seconds.
- */
-#define MINIMUM_SLEEP_INTERVAL    5
+#include "status.h"
+#include "sysfs.h"
+#include "udev.h"
+#include "utils.h"
+#include "version.h"
+#include "vmdssd.h"
 
 /**
  * @brief List of active block devices.
@@ -71,7 +65,7 @@
  * Only devices which have enclosure management feature enabled are on the
  * list, other devices are ignored (except protocol is forced).
  */
-static void *ledmon_block_list = NULL;
+static struct list ledmon_block_list;
 
 /**
  * @brief Daemon process termination flag.
@@ -79,17 +73,16 @@ static void *ledmon_block_list = NULL;
  * This flag indicates that daemon process should terminate. User must send
  * SIGTERM to daemon in order to terminate the process gently.
  */
-static sig_atomic_t terminate = 0;
+static sig_atomic_t terminate;
 
 /**
- * @brief The interval between sysfs scans.
+ * @brief Path to ledmon configuration file.
  *
- * This static variable holds the value of time interval used by the daemon to
- * set sleep between two scans of sysfs. The time interval is given in seconds.
- * The value is set to DEFAULT_SLEEP_INTERVAL by default and it can be change
- * with appropriate command line option. The minimum time interval is 5 seconds.
+ * This string contains path of the ledmon configuration file. The value is
+ * set to LEDMON_DEF_CONF_FILE by default and it can be changed by command line
+ * option.
  */
-static int sleep_interval = DEFAULT_SLEEP_INTERVAL;
+static char *ledmon_conf_path;
 
 /**
  * @brief Name of IBPI patterns.
@@ -103,13 +96,14 @@ const char *ibpi_str[] = {
 	[IBPI_PATTERN_ONESHOT_NORMAL] = "Oneshot Off",
 	[IBPI_PATTERN_DEGRADED]       = "In a Critical Array",
 	[IBPI_PATTERN_REBUILD]        = "Rebuild",
-	[IBPI_PATTERN_REBUILD_P]      = "Rebuild Preferred",
 	[IBPI_PATTERN_FAILED_ARRAY]   = "In a Failed Array",
 	[IBPI_PATTERN_HOTSPARE]       = "Hotspare",
 	[IBPI_PATTERN_PFA]            = "Predicted Failure Analysis",
 	[IBPI_PATTERN_FAILED_DRIVE]   = "Failure",
 	[IBPI_PATTERN_LOCATE]         = "Locate",
-	[IBPI_PATTERN_LOCATE_OFF]     = "Locate Off"
+	[IBPI_PATTERN_LOCATE_OFF]     = "Locate Off",
+	[IBPI_PATTERN_ADDED]          = "Added",
+	[IBPI_PATTERN_REMOVED]        = "Removed"
 };
 
 /**
@@ -117,7 +111,7 @@ const char *ibpi_str[] = {
  * information about the version of monitor service.
  */
 static char *ledmon_version = "Intel(R) Enclosure LED Monitor Service %d.%d\n"
-			      "Copyright (C) 2009-2013 Intel Corporation.\n";
+			      "Copyright (C) 2009-2018 Intel Corporation.\n";
 
 /**
  * Internal variable of monitor service. It is used to help parse command line
@@ -139,7 +133,8 @@ enum longopt {
 	OPT_LOG,
 	OPT_QUIET,
 	OPT_VERSION,
-	OPT_WARNING
+	OPT_WARNING,
+	OPT_LOG_LEVEL,
 };
 
 /**
@@ -158,6 +153,7 @@ static struct option longopt[] = {
 	[OPT_QUIET]    = {"quiet", no_argument, NULL, '\0'},
 	[OPT_VERSION]  = {"version", no_argument, NULL, 'v'},
 	[OPT_WARNING]  = {"warning", no_argument, NULL, '\0'},
+	[OPT_LOG_LEVEL] = {"log-level", required_argument, NULL, '\0'},
 			 {NULL, no_argument, NULL, '\0'}
 };
 
@@ -176,11 +172,8 @@ static struct option longopt[] = {
  */
 static void _ledmon_fini(int __attribute__ ((unused)) status, void *progname)
 {
-	sysfs_fini();
-	if (ledmon_block_list) {
-		list_for_each(ledmon_block_list, block_device_fini);
-		list_fini(ledmon_block_list);
-	}
+	sysfs_reset();
+	list_erase(&ledmon_block_list);
 	log_close();
 	pidfile_remove(progname);
 }
@@ -242,19 +235,20 @@ static void _ledmon_help(void)
 	printf("\nUsage: %s [OPTIONS]\n\n", progname);
 	printf("Mandatory arguments for long options are mandatory for short "
 	       "options, too.\n\n");
-	printf("--interval=VALUE\t\t  Set time interval to VALUE seconds.\n"
-	       "\t\t\t\t  The smallest interval is 5 seconds.\n");
-	printf("--config=PATH,  -c PATH\t\t  Use alternate configuration file "
-	       "(not yet\n\t\t\t\t  implemented).\n");
-	printf
-	    ("--log=PATH\t\t\t  Use local log file instead\n\t\t\t\t  /var/log/"
-	     "ledmon.log global file.\n");
-	printf("--help\t\t\t\t  Displays this help text.\n");
-	printf
-	    ("--version\t\t\t  Displays version and license information.\n\n");
-	printf("Refer to ledmon(8) man page for more detailed description.\n");
-	printf("Bugs should be reported at: "
-	       "http://sourceforge.net/p/ledmon/bugs \n");
+	print_opt("--interval=VALUE", "-t VALUE",
+			  "Set time interval to VALUE seconds.");
+	print_opt("", "", "The smallest interval is 5 seconds.");
+	print_opt("--config=PATH", "-c PATH",
+			  "Use alternate configuration file.");
+	print_opt("--log=PATH", "-l PATH",
+			  "Use local log file instead /var/log/ledmon.log");
+	print_opt("--log-level=VALUE", "-l VALUE",
+			  "Allows user to set ledmon verbose level in logs.");
+	print_opt("--help", "-h", "Displays this help text.");
+	print_opt("--version", "-v",
+			  "Displays version and license information.");
+	printf("\nRefer to ledmon(8) man page for more detailed description.\n");
+	printf("Bugs should be reported at: https://github.com/intel/ledmon/issues\n");
 }
 
 /**
@@ -268,38 +262,15 @@ static void _ledmon_help(void)
  *
  * @return STATUS_SUCCESS if successful, otherwise a valid status_t status code.
  */
-static status_t _set_config_path(const char *path)
+static status_t _set_config_path(char **conf_path, const char *path)
 {
-	(void)path;
-	return STATUS_SUCCESS;
-}
+	if (!path)
+		path = LEDMON_DEF_CONF_FILE;
 
-/**
- * @brief Sets the path to local log file.
- *
- * This is internal function of monitor service. This function sets the path and
- * file name of log file. The function checks if the specified path is valid. In
- * case of incorrect path the function does nothing.
- *
- * @param[in]      path           new location and name of log file.
- *
- * @return STATUS_SUCCESS if successful, otherwise a valid status_t status code.
- *         The following status code are returned:
- *
- *         STATUS_INVALID_PATH    the given path is invalid.
- *         STATUS_FILE_OPEN_ERROR unable to open a log file i.e. because of
- *                                insufficient privileges.
- */
-static status_t _set_log_path(const char *path)
-{
-	char temp[PATH_MAX];
+	if (*conf_path)
+		free(*conf_path);
+	*conf_path = strdup(path);
 
-	if (realpath(path, temp) == NULL) {
-		if ((errno != ENOENT) && (errno != ENOTDIR))
-			return STATUS_INVALID_PATH;
-	}
-	if (log_open(temp) < 0)
-		return STATUS_FILE_OPEN_ERROR;
 	return STATUS_SUCCESS;
 }
 
@@ -316,14 +287,122 @@ static status_t _set_log_path(const char *path)
  */
 static status_t _set_sleep_interval(const char *optarg)
 {
-	sleep_interval = atoi(optarg);
-	if (sleep_interval < MINIMUM_SLEEP_INTERVAL) {
+	conf.scan_interval = atoi(optarg);
+	if (conf.scan_interval < LEDMON_MIN_SLEEP_INTERVAL) {
 		log_warning("sleep interval too small... using default.");
-		sleep_interval = DEFAULT_SLEEP_INTERVAL;
+		conf.scan_interval = LEDMON_DEF_SLEEP_INTERVAL;
 	}
 	return STATUS_SUCCESS;
 }
 
+/**
+ * @brief Sets verbose variable to given level.
+ *
+ * This is internal function of monitor service. The function maps given level
+ * to the value from verbose_level enum and sets verbose value to ledmon
+ * configuration.
+ *
+ * @param[in]      log_level     required new log_level.
+ *
+ * @return STATUS_SUCCESS if successful, otherwise a valid status_t status code.
+ */
+static status_t _set_verbose_level(int log_level)
+{
+	int new_verbose = -1;
+
+	switch (log_level) {
+	case OPT_ALL:
+		new_verbose = LOG_LEVEL_ALL;
+		break;
+	case OPT_DEBUG:
+		new_verbose = LOG_LEVEL_DEBUG;
+		break;
+	case OPT_ERROR:
+		new_verbose = LOG_LEVEL_ERROR;
+		break;
+	case OPT_INFO:
+		new_verbose = LOG_LEVEL_INFO;
+		break;
+	case OPT_QUIET:
+		new_verbose = LOG_LEVEL_QUIET;
+		break;
+	case OPT_WARNING:
+		new_verbose = LOG_LEVEL_WARNING;
+		break;
+	}
+	if (new_verbose != -1) {
+		conf.log_level = new_verbose;
+		return STATUS_SUCCESS;
+	}
+	return STATUS_CMDLINE_ERROR;
+}
+
+/**
+ * @brief Gets id for given CLI option which corresponds to value from longopt
+ * table.
+ *
+ * This is internal function of monitor service. The function maps given string
+ * to the value from longopt enum and returns id of matched element. Generic
+ * parameters allow to use this function for any CLI options-table which bases
+ * on option struct.
+ *
+ * @param[in]     optarg          String containing value given by user in CLI.
+ * @param[in]     longopt         Table of allowed CLI options.
+ *
+ * @return integer id if successful, otherwise a -1.
+ */
+static int _get_option_id(const char *optarg, struct option longopt[])
+{
+	struct option *i = longopt;
+
+	while (i->name != NULL) {
+		if (strcmp(i->name, optarg) == 0)
+			return i-longopt;
+		i++;
+	}
+	return -1;
+}
+
+/**
+ * @brief Reads config file path and checks if command line input contains
+ * options which don't require to run ledmon as daemon.
+ *
+ * This is internal function of monitor service. This function looks for
+ * config file path in command line options given to the program from
+ * command line interface. It also handles options to print help and version.
+ *
+ * @param[in]     argc            - number of arguments.
+ * @param[in]     argv            - array of command line arguments.
+ *
+ * @return STATUS_SUCCESS if successful, otherwise a valid status_t status code.
+ */
+static status_t _cmdline_parse_non_daemonise(int argc, char *argv[])
+{
+	int opt_index = -1;
+	int opt = -1;
+	status_t status = STATUS_SUCCESS;
+
+	do {
+		opt = getopt_long(argc, argv, shortopt, longopt, &opt_index);
+		switch (opt) {
+		case 'c':
+			status = _set_config_path(&ledmon_conf_path, optarg);
+			break;
+		case 'h':
+			_ledmon_help();
+			exit(EXIT_SUCCESS);
+		case 'v':
+			_ledmon_version();
+			exit(EXIT_SUCCESS);
+		case ':':
+		case '?':
+			return STATUS_CMDLINE_ERROR;
+		}
+	} while (opt >= 0);
+
+	return status;
+}
+
 /**
  * @brief Command line interface handler function.
  *
@@ -340,53 +419,34 @@ static status_t _cmdline_parse(int argc, char *argv[])
 	int opt, opt_index = -1;
 	status_t status = STATUS_SUCCESS;
 
+	optind = 1;
 	do {
 		opt = getopt_long(argc, argv, shortopt, longopt, &opt_index);
 		if (opt == -1)
 			break;
+		if (opt == 'c')
+			continue;
 		switch (opt) {
+		int new_log_level;
 		case 0:
 			switch (opt_index) {
-			case OPT_ALL:
-				verbose = VERB_ALL;
-				break;
-			case OPT_DEBUG:
-				verbose = VERB_DEBUG;
-				break;
-			case OPT_ERROR:
-				verbose = VERB_ERROR;
-				break;
-			case OPT_INFO:
-				verbose = VERB_INFO;
-				break;
-			case OPT_QUIET:
-				verbose = VERB_QUIET;
-				break;
-			case OPT_WARNING:
-				verbose = VERB_WARN;
+			case OPT_LOG_LEVEL:
+				new_log_level = _get_option_id(optarg, longopt);
+				if (new_log_level != -1)
+					status = _set_verbose_level(new_log_level);
+				else
+					status = STATUS_CMDLINE_ERROR;
 				break;
+			default:
+				status = _set_verbose_level(opt);
 			}
 			break;
 		case 'l':
-			status = _set_log_path(optarg);
+			status = set_log_path(optarg);
 			break;
 		case 't':
 			status = _set_sleep_interval(optarg);
 			break;
-		case 'c':
-			status = _set_config_path(optarg);
-			break;
-		case 'v':
-			_ledmon_version();
-			exit(EXIT_SUCCESS);
-		case 'h':
-			_ledmon_help();
-			exit(EXIT_SUCCESS);
-		case ':':
-		case '?':
-		default:
-			log_debug("[opt='%c', opt_index=%d]", opt, opt_index);
-			break;
 		}
 		opt_index = -1;
 		if (status != STATUS_SUCCESS)
@@ -463,8 +523,8 @@ static void _ledmon_setup_signals(void)
  */
 static void _ledmon_wait(int seconds)
 {
-	fd_set rd;
-	int fd;
+	int fd, udev_fd, max_fd, res;
+	fd_set rdfds, exfds;
 	struct timespec timeout;
 	sigset_t sigset;
 
@@ -472,90 +532,152 @@ static void _ledmon_wait(int seconds)
 	sigdelset(&sigset, SIGTERM);
 	timeout.tv_nsec = 0;
 	timeout.tv_sec = seconds;
-	FD_ZERO(&rd);
+
 	fd = open("/proc/mdstat", O_RDONLY);
-	if (fd)
-		FD_SET(fd, &rd);
-	pselect(fd + 1, NULL, NULL, &rd, &timeout, &sigset);
+	udev_fd = get_udev_monitor();
+	max_fd = MAX(fd, udev_fd) + 1;
+	do {
+		FD_ZERO(&rdfds);
+		FD_ZERO(&exfds);
+
+		if (fd > 0)
+			FD_SET(fd, &exfds);
+		if (udev_fd > 0)
+			FD_SET(udev_fd, &rdfds);
+
+		res = pselect(max_fd, &rdfds, NULL, &exfds, &timeout, &sigset);
+		if (!FD_ISSET(udev_fd, &rdfds) ||
+		    handle_udev_event(&ledmon_block_list) <= 0)
+			break;
+	} while (res > 0);
+
 	if (fd >= 0)
 		close(fd);
 }
 
-static int is_dellssd(struct block_device *bd)
-{
-	return (bd->cntrl && bd->cntrl->cntrl_type == CNTRL_TYPE_DELLSSD);
-}
-
 /**
- * @brief Checks the presence of block device.
+ * @brief Determine failed state by comparing saved block device with new
+ * scanned.
  *
- * This is internal function of monitor service. The function is checking
- * whether block device is already on the list or it is missing from the list.
- * The function is design to be used as 'test' parameter for list_find_first()
- * function.
+ * This is internal function of monitor service. Due race conditions related
+ * with removing files from /sys/block/md* when raid is stopped or disk is
+ * failed, this function analyse state of every block device between scans.
  *
- * @param[in]    blk1            - an element from a list to compare to.
- * @param[in]    blk2            - a block device being searched.
+ * @param[in]    block           Pointer to new (scanned) block device
+ *				 structure.
+ * @param[in]    temp            Pointer to previously saved state of block
+ *				 device structure.
  *
- * @return 0 if the block devices do not match, otherwise function returns 1.
+ * @return The function does not return a value.
  */
-static int _compare(struct block_device *bd_old, struct block_device *bd_new)
+static void _handle_fail_state(struct block_device *block,
+			       struct block_device *temp)
 {
-	int i = 0;
+	struct raid_device *temp_raid_device = NULL;
 
-	if (!is_dellssd(bd_old) && bd_old->host_id == -1) {
-		log_debug("Device %s : No host_id!",
-			  strstr(bd_old->sysfs_path, "host"));
-		return 0;
-	}
-	if (!is_dellssd(bd_new) && bd_new->host_id == -1) {
-		log_debug("Device %s : No host_id!",
-			  strstr(bd_new->sysfs_path, "host"));
-		return 0;
-	}
-	if (!bd_old->cntrl) {
-		log_debug("Device %s : No ctrl dev!",
-			  strstr(bd_old->sysfs_path, "host"));
-		return 0;
-	}
+	if (!temp->raid_dev)
+		/*
+		 * Device is a RAID member now, so keep information about
+		 * related with device RAID.
+		 */
+		temp->raid_dev = raid_device_duplicate(block->raid_dev);
 
-	if (bd_old->cntrl->cntrl_type != bd_new->cntrl->cntrl_type)
-		return 0;
+	if (!temp->raid_dev)
+		return;
 
-	switch (bd_old->cntrl->cntrl_type) {
-	case CNTRL_TYPE_AHCI:
-		/* Missing support for port multipliers. Compare just hostX. */
-		i = (bd_old->host_id == bd_new->host_id);
-		break;
+	temp_raid_device = find_raid_device(sysfs_get_volumes(),
+					    temp->raid_dev->sysfs_path);
 
-	case CNTRL_TYPE_SCSI:
-		/* Host and phy is not enough. They might be DA or EA. */
-		if (dev_directly_attached(bd_old->sysfs_path) &&
-		    dev_directly_attached(bd_new->sysfs_path)) {
-			/* Just compare host & phy */
-			i = (bd_old->host_id == bd_new->host_id) &&
-			    (bd_old->phy_index == bd_new->phy_index);
-
-			break;
+	if (!block->raid_dev) {
+		if (temp->raid_dev->type == DEVICE_TYPE_VOLUME &&
+		    temp_raid_device) {
+			/*
+			 * Device is outside of the volume, but related raid
+			 * still exist, so disk has been removed from volume -
+			 * blink fail LED. It is case when drive is removed
+			 * by mdadm -If.
+			 */
+			temp->ibpi = IBPI_PATTERN_FAILED_DRIVE;
+			/*
+			 * Changing failed state to hotspare will be prevent by
+			 * code from _add_block function. If disk come back to
+			 * container failed state should be removed. By setting
+			 * type to CONTAINER ledmon can react in this case.
+			 */
+			temp->raid_dev->type = DEVICE_TYPE_CONTAINER;
+		} else {
+			/*
+			 * Device was RAID member and was failed (was outside
+			 * of array and container). Now again is in container.
+			 * Release object to perform hotspare state.
+			 * Or:
+			 * Device is outside of the volume, but related raid is
+			 * removed (or stopped) so device is no longer a RAID
+			 * member.
+			 */
+			raid_device_fini(temp->raid_dev);
+			temp->raid_dev = NULL;
 		}
-		if (!dev_directly_attached(bd_old->sysfs_path) &&
-		    !dev_directly_attached(bd_new->sysfs_path)) {
-			/* Both expander attached */
-			i = (bd_old->host_id == bd_new->host_id) &&
-			    (bd_old->phy_index == bd_new->phy_index);
-			i = i && (bd_old->encl_index == bd_new->encl_index);
-			break;
-		}
-		/* */
-		break;
+	} else if (block->raid_dev) {
+		if (temp->raid_dev->type == DEVICE_TYPE_VOLUME &&
+		    block->raid_dev->type == DEVICE_TYPE_CONTAINER) {
+			/*
+			 * Drive is removed from volume, but still exist
+			 * in container.
+			 */
+			enum raid_level new_level;
 
-	case CNTRL_TYPE_DELLSSD:
-	default:
-		/* Just compare names */
-		i = (strcmp(bd_old->sysfs_path, bd_new->sysfs_path) == 0);
-		break;
+			if (!temp_raid_device)
+				new_level = RAID_LEVEL_UNKNOWN;
+			else
+				new_level = temp_raid_device->level;
+
+			if ((temp->raid_dev->level == RAID_LEVEL_10 ||
+				 temp->raid_dev->level == RAID_LEVEL_1) &&
+				 new_level == RAID_LEVEL_0) {
+				/*
+				 * Device is removed from volume due to
+				 * migration to raid. State of this disk is
+				 * hotspare now.
+				 */
+				temp->ibpi = IBPI_PATTERN_HOTSPARE;
+			} else {
+				/*
+				 * Trasitions other than raid 0 migration.
+				 * Like reshape, volume stopping etc.
+				 */
+				if (!temp_raid_device) {
+					if (block->raid_dev->sync_action !=
+						RAID_ACTION_RESHAPE)
+						/*
+						 * Fail state should be
+						 * released,because disk is in
+						 * reshape process.
+						 */
+						temp->ibpi =
+							IBPI_PATTERN_HOTSPARE;
+				} else {
+					/*
+					 * Drive is removed from volume,
+					 * but still exist in container. This
+					 * situation can be caused by bad
+					 * blocks or calling mdadm
+					 * --set-faulty.
+					 */
+					temp->ibpi = IBPI_PATTERN_FAILED_DRIVE;
+				}
+			}
+		} else if (temp->raid_dev->type == DEVICE_TYPE_CONTAINER &&
+			   block->raid_dev->type == DEVICE_TYPE_VOLUME) {
+			/*
+			 * Disk was in container and is added to volume.
+			 * Release object for recreating.
+			 */
+			raid_device_fini(temp->raid_dev);
+			temp->raid_dev =
+				raid_device_duplicate(block->raid_dev);
+		}
 	}
-	return i;
 }
 
 /**
@@ -574,13 +696,19 @@ static int _compare(struct block_device *bd_old, struct block_device *bd_new)
  */
 static void _add_block(struct block_device *block)
 {
-	struct block_device *temp;
+	struct block_device *temp = NULL;
 
-	temp = list_first_that(ledmon_block_list, _compare, block);
+	list_for_each(&ledmon_block_list, temp) {
+		if (block_compare(temp, block))
+			break;
+		temp = NULL;
+	}
 	if (temp) {
 		enum ibpi_pattern ibpi = temp->ibpi;
 		temp->timestamp = block->timestamp;
-		if (temp->ibpi == IBPI_PATTERN_ONESHOT_NORMAL) {
+		if (temp->ibpi == IBPI_PATTERN_ADDED) {
+			temp->ibpi = IBPI_PATTERN_ONESHOT_NORMAL;
+		} else if (temp->ibpi == IBPI_PATTERN_ONESHOT_NORMAL) {
 			temp->ibpi = IBPI_PATTERN_UNKNOWN;
 		} else if (temp->ibpi != IBPI_PATTERN_FAILED_DRIVE) {
 			if (block->ibpi == IBPI_PATTERN_UNKNOWN) {
@@ -594,21 +722,24 @@ static void _add_block(struct block_device *block)
 			} else {
 				temp->ibpi = block->ibpi;
 			}
-		} else {
-			temp->ibpi = IBPI_PATTERN_ONESHOT_NORMAL;
+		} else if (!(temp->ibpi == IBPI_PATTERN_FAILED_DRIVE &&
+			block->ibpi == IBPI_PATTERN_HOTSPARE) ||
+			(temp->ibpi == IBPI_PATTERN_FAILED_DRIVE &&
+			block->ibpi == IBPI_PATTERN_NONE)) {
+			temp->ibpi = block->ibpi;
 		}
 
-		if (ibpi != temp->ibpi) {
+		_handle_fail_state(block, temp);
+
+		if (ibpi != temp->ibpi && ibpi <= IBPI_PATTERN_REMOVED) {
 			log_info("CHANGE %s: from '%s' to '%s'.",
 				 temp->sysfs_path, ibpi_str[ibpi],
 				 ibpi_str[temp->ibpi]);
 		}
 		/* Check if name of the device changed.*/
-		/* It's possible for SCSI devices.     */
 		if (strcmp(temp->sysfs_path, block->sysfs_path)) {
 			log_info("NAME CHANGED %s to %s",
-				 strstr(temp->sysfs_path, "host"),
-				 strstr(block->sysfs_path, "host"));
+				 temp->sysfs_path, block->sysfs_path);
 			free(temp->sysfs_path);
 			temp->sysfs_path = strdup(block->sysfs_path);
 		}
@@ -618,9 +749,7 @@ static void _add_block(struct block_device *block)
 		if (temp != NULL) {
 			log_info("NEW %s: state '%s'.", temp->sysfs_path,
 				 ibpi_str[temp->ibpi]);
-			list_put(ledmon_block_list, temp,
-				 sizeof(struct block_device));
-			free(temp);
+			list_append(&ledmon_block_list, temp);
 		}
 	}
 }
@@ -647,15 +776,17 @@ static void _send_msg(struct block_device *block)
 			  strstr(block->sysfs_path, "host"));
 		return;
 	}
-	if (block->timestamp != timestamp) {
+	if (block->timestamp != timestamp ||
+	    block->ibpi == IBPI_PATTERN_REMOVED) {
 		if (block->ibpi != IBPI_PATTERN_FAILED_DRIVE) {
 			log_info("CHANGE %s: from '%s' to '%s'.",
 				 block->sysfs_path, ibpi_str[block->ibpi],
 				 ibpi_str[IBPI_PATTERN_FAILED_DRIVE]);
 			block->ibpi = IBPI_PATTERN_FAILED_DRIVE;
 		} else {
+			char *host = strstr(block->sysfs_path, "host");
 			log_debug("DETACHED DEV '%s' in failed state",
-				  strstr(block->sysfs_path, "host"));
+				  host ? host : block->sysfs_path);
 		}
 	}
 	block->send_fn(block, block->ibpi);
@@ -675,19 +806,22 @@ static void _revalidate_dev(struct block_device *block)
 	block->cntrl = block_get_controller(sysfs_get_cntrl_devices(),
 					    block->cntrl_path);
 	if (!block->cntrl) {
+		/* It could be removed VMD drive */
 		log_debug("Failed to get controller for dev: %s, ctrl path: %s",
 			  block->sysfs_path, block->cntrl_path);
 		return;
 	}
-	if (block->cntrl->cntrl_type == CNTRL_TYPE_SCSI &&
-	    block->cntrl->isci_present) {
+	if (block->cntrl->cntrl_type == CNTRL_TYPE_SCSI) {
 		block->host = block_get_host(block->cntrl, block->host_id);
 		if (block->host) {
-			isci_cntrl_init_smp(NULL, block->cntrl);
+			if (dev_directly_attached(block->sysfs_path))
+				cntrl_init_smp(NULL, block->cntrl);
+			else
+				scsi_get_enclosure(block);
 		} else {
 			log_debug("Failed to get host for dev: %s, hostId: %d",
 				  block->sysfs_path, block->host_id);
-			/* If hosts failed for isci, invalidate cntrl */
+			/* If failed, invalidate cntrl */
 			block->cntrl = NULL;
 		}
 	}
@@ -699,24 +833,15 @@ static void _invalidate_dev(struct block_device *block)
 	/* Those fields are valid only per 'session' - through single scan. */
 	block->cntrl = NULL;
 	block->host = NULL;
+	block->enclosure = NULL;
+	block->encl_index = -1;
 }
 
 static void _check_block_dev(struct block_device *block, int *restart)
 {
-	/* Check SCSI device behind expander. */
-	if (block->cntrl->cntrl_type == CNTRL_TYPE_SCSI) {
-		if (dev_directly_attached(block->sysfs_path) == 0) {
-			if (block->ibpi == IBPI_PATTERN_FAILED_DRIVE &&
-			    (block->encl_index == -1
-			     || block->encl_dev[0] == 0)) {
-				(*restart)++;
-				log_debug("%s(): invalidating device: %s. "
-					"No link to enclosure", __func__,
-					strstr(block->sysfs_path, "host"));
-			}
-		}
+	if (!block->cntrl) {
+		(*restart)++;
 	}
-	return;
 }
 
 /**
@@ -733,31 +858,60 @@ static void _check_block_dev(struct block_device *block, int *restart)
 static void _ledmon_execute(void)
 {
 	int restart = 0;	/* ledmon_block_list needs restart? */
-	status_t status = STATUS_SUCCESS;
+	struct block_device *device;
 
 	/* Revalidate each device in the list. Bring back controller and host */
-	list_for_each(ledmon_block_list, _revalidate_dev);
+	list_for_each(&ledmon_block_list, device)
+		_revalidate_dev(device);
 	/* Scan all devices and compare them against saved list */
-	sysfs_block_device_for_each(_add_block);
+	list_for_each(sysfs_get_block_devices(), device)
+		_add_block(device);
 	/* Send message to all devices in the list if needed. */
-	list_for_each(ledmon_block_list, _send_msg);
+	list_for_each(&ledmon_block_list, device)
+		_send_msg(device);
 	/* Flush unsent messages from internal buffers. */
-	list_for_each(ledmon_block_list, _flush_msg);
+	list_for_each(&ledmon_block_list, device)
+		_flush_msg(device);
 	/* Check if there is any orphaned device. */
-	list_for_each_parm(ledmon_block_list, _check_block_dev, &restart);
-	/* Invalidate each device in the list. Clear controller and host. */
-	list_for_each(ledmon_block_list, _invalidate_dev);
+	list_for_each(&ledmon_block_list, device)
+		_check_block_dev(device, &restart);
 
 	if (restart) {
 		/* there is at least one detached element in the list. */
-		list_for_each(ledmon_block_list, block_device_fini);
-		list_fini(ledmon_block_list);
-		status = list_init(&ledmon_block_list);
-		if (status != STATUS_SUCCESS) {
-			log_debug("%s(): list_init() failed (status=%s).",
-				  __func__, strstatus(status));
-			exit(EXIT_FAILURE);
+		list_erase(&ledmon_block_list);
+	}
+}
+
+static status_t _init_ledmon_conf(void)
+{
+	memset(&conf, 0, sizeof(struct ledmon_conf));
+
+	/* initialize with default values */
+	conf.blink_on_init = 1;
+	conf.blink_on_migration = 1;
+	conf.rebuild_blink_on_all = 0;
+	conf.raid_memebers_only = 0;
+	conf.log_level = LOG_LEVEL_WARNING;
+	conf.scan_interval = LEDMON_DEF_SLEEP_INTERVAL;
+	list_init(&conf.cntrls_whitelist, NULL);
+	list_init(&conf.cntrls_blacklist, NULL);
+	return set_log_path(LEDMON_DEF_LOG_FILE);
+}
+
+static void _close_parent_fds(void)
+{
+	struct list dir;
+
+	if (scan_dir("/proc/self/fd", &dir) == 0) {
+		char *elem;
+
+		list_for_each(&dir, elem) {
+			int fd = (int)strtol(basename(elem), NULL, 10);
+
+			if (fd != get_log_fd())
+				close(fd);
 		}
+		list_erase(&dir);
 	}
 }
 
@@ -766,18 +920,32 @@ static void _ledmon_execute(void)
 int main(int argc, char *argv[])
 {
 	status_t status = STATUS_SUCCESS;
-	int i;
 
 	set_invocation_name(argv[0]);
 	openlog(progname, LOG_PID | LOG_PERROR, LOG_DAEMON);
 
+	if (_cmdline_parse_non_daemonise(argc, argv) != STATUS_SUCCESS)
+		return STATUS_CMDLINE_ERROR;
+
 	if (getuid() != 0) {
-		log_error("Only root can run this application.");
+		fprintf(stderr, "Only root can run this application.\n");
 		return STATUS_NOT_A_PRIVILEGED_USER;
 	}
 
+	status = _init_ledmon_conf();
+	if (status != STATUS_SUCCESS)
+		return status;
+
 	if (on_exit(_ledmon_status, &terminate))
 		return STATUS_ONEXIT_ERROR;
+
+	status = ledmon_read_config(ledmon_conf_path);
+	if (status != STATUS_SUCCESS)
+		return status;
+
+	if(conf.log_path)
+		set_log_path(conf.log_path);
+
 	if (_cmdline_parse(argc, argv) != STATUS_SUCCESS)
 		return STATUS_CMDLINE_ERROR;
 
@@ -799,8 +967,9 @@ int main(int argc, char *argv[])
 		log_debug("main(): setsid() failed (errno=%d).", errno);
 		exit(EXIT_FAILURE);
 	}
-	for (i = getdtablesize() - 1; i >= 0; --i)
-		close(i);
+
+	_close_parent_fds();
+
 	int t = open("/dev/null", O_RDWR);
 	dup(t);
 	dup(t);
@@ -818,34 +987,21 @@ int main(int argc, char *argv[])
 
 	if (on_exit(_ledmon_fini, progname))
 		exit(STATUS_ONEXIT_ERROR);
-	status = list_init(&ledmon_block_list);
-	if (status != STATUS_SUCCESS) {
-		log_debug("main(): list_init() failed (status=%s).",
-			  strstatus(status));
-		exit(EXIT_FAILURE);
-	}
-	status = sysfs_init();
-	if (status != STATUS_SUCCESS) {
-		log_debug("main(): sysfs_init() failed (status=%s).",
-			  strstatus(status));
-		exit(EXIT_FAILURE);
-	}
+	list_init(&ledmon_block_list, (item_free_t)block_device_fini);
+	sysfs_init();
 	log_info("monitor service has been started...");
 	while (terminate == 0) {
+		struct block_device *device;
+
 		timestamp = time(NULL);
-		status = sysfs_scan();
-		if (status != STATUS_SUCCESS) {
-			log_debug("main(): sysfs_scan() failed (status=%s).",
-				  strstatus(status));
-		} else {
-			_ledmon_execute();
-			status = sysfs_reset();
-			if (status != STATUS_SUCCESS) {
-				log_debug("main(): sysfs_reset() failed "
-					  "(status=%s).", strstatus(status));
-			}
-		}
-		_ledmon_wait(sleep_interval);
+		sysfs_scan();
+		_ledmon_execute();
+		_ledmon_wait(conf.scan_interval);
+		/* Invalidate each device in the list. Clear controller and host. */
+		list_for_each(&ledmon_block_list, device)
+			_invalidate_dev(device);
+		sysfs_reset();
 	}
+	stop_udev_monitor();
 	exit(EXIT_SUCCESS);
 }
