@@ -1,6 +1,6 @@
 /*
  * Intel(R) Enclosure LED Utilities
- * Copyright (C) 2009-2013 Intel Corporation.
+ * Copyright (C) 2009-2018 Intel Corporation.
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms and conditions of the GNU General Public License,
@@ -17,18 +17,16 @@
  *
  */
 
-#include <config.h>
-
-#include <sys/stat.h>
+#include <dirent.h>
+#include <errno.h>
 #include <fcntl.h>
 #include <limits.h>
-#include <errno.h>
-#include <unistd.h>
+#include <stdint.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <stdio.h>
-#include <stdint.h>
-#include <dirent.h>
+#include <sys/stat.h>
+#include <unistd.h>
 
 #if _HAVE_DMALLOC_H
 #include <dmalloc.h>
@@ -37,52 +35,30 @@
 #include <scsi/sg_lib.h>
 #include <scsi/sg_cmds_extra.h>
 
-#include "ibpi.h"
-#include "status.h"
-#include "list.h"
-#include "block.h"
 #include "cntrl.h"
-#include "scsi.h"
+#include "config.h"
 #include "enclosure.h"
+#include "list.h"
+#include "scsi.h"
+#include "ses.h"
+#include "status.h"
 #include "sysfs.h"
 #include "utils.h"
-#include "ses.h"
 
 static int debug = 0;
 
-static void print_page10(struct ses_pages *);
-
-static void send_diag_slot(struct ses_pages *, int, const unsigned char *, int);
-
-static int ses_set_message(enum ibpi_pattern, unsigned char *);
-
-static inline int get_page2(int fd, unsigned char *p2, int *len)
+static int get_ses_page(int fd, struct ses_page *p, int pg_code)
 {
 	int ret;
-	/* Get Enclosure Status */
-	ret = sg_ll_receive_diag(fd, 1, ENCL_CTRL_DIAG_STATUS,
-				 p2, SES_ALLOC_BUFF, 0, debug);
-	*len = (p2[2] << 8) + p2[3] + 4;
-	return ret;
-}
+	int retry_count = 3;
 
-static inline int get_page1(int fd, unsigned char *p1, int *len)
-{
-	int ret;
-	/* Get Enclosure Status */
-	ret = sg_ll_receive_diag(fd, 1, ENCL_CFG_DIAG_STATUS,
-				 p1, SES_ALLOC_BUFF, 0, debug);
-	*len = (p1[2] << 8) + p1[3] + 4;
-	return ret;
-}
+	do {
+		ret = sg_ll_receive_diag(fd, 1, pg_code, p->buf, sizeof(p->buf),
+					 0, debug);
+	} while (ret && retry_count--);
 
-static inline int get_page10(int fd, unsigned char *p10, int *len)
-{
-	int ret;
-	/* Get Enclosure Status */
-	ret = sg_ll_receive_diag(fd, 1, ENCL_ADDITIONAL_EL_STATUS,
-				 p10, SES_ALLOC_BUFF, 0, debug);
-	*len = (p10[2] << 8) + p10[3] + 4;
+	if (!ret)
+		p->len = (p->buf[2] << 8) + p->buf[3] + 4;
 	return ret;
 }
 
@@ -93,15 +69,13 @@ static int process_page1(struct ses_pages *sp)
 	int len = 0;
 	int sum_headers = 0;	/* Number of Type descriptor headers */
 	int i = 0;
-	int components = 0;	/* Total number of components in enclosure */
 
 	/* How many enclosures is in the main enclosure? */
-	num_encl = sp->page1[1] + 1;
-	sp->page1_len = (sp->page1[2] << 8) + sp->page1[3] + 4;
+	num_encl = sp->page1->buf[1] + 1;
 	/* Go to Enclosure Descriptor */
-	ed = sp->page1 + 8;
+	ed = sp->page1->buf + 8;
 	for (i = 0; i < num_encl; i++, ed += len) {
-		if (ed + 3 > sp->page1 + sp->page1_len) {
+		if (ed + 3 > sp->page1->buf + sp->page1->len) {
 			log_debug
 			    ("SES: Error, response pare 1 truncated at %d\n",
 			     i);
@@ -115,19 +89,16 @@ static int process_page1(struct ses_pages *sp)
 		}
 	}
 
-	sp->page1_types = ed;
+	sp->page1_types = (struct type_descriptor_header *)ed;
 	sp->page1_types_len = sum_headers;
 
 	/* ed is on type descr header */
 	for (i = 0; i < sum_headers; i++, ed += 4) {
-		if (ed > sp->page1 + sp->page1_len) {
+		if (ed > sp->page1->buf + sp->page1->len) {
 			log_debug("SES: Response page 1 truncated at %d\n", i);
 			return 1;
 		}
-		if (ed[0] == SES_DEVICE_SLOT || ed[0] == SES_ARRAY_DEVICE_SLOT)
-			components += ed[1];
 	}
-	sp->components = components;
 	return 0;
 }
 
@@ -137,19 +108,13 @@ static struct ses_pages *ses_init(void)
 	sp = calloc(1, sizeof(*sp));
 	if (!sp)
 		return NULL;
-	sp->page1 = calloc(SES_ALLOC_BUFF, sizeof(*sp->page1));
+	sp->page1 = calloc(1, sizeof(struct ses_page));
 	if (!sp->page1)
 		goto sp1;
-	sp->page2 = calloc(SES_ALLOC_BUFF, sizeof(*sp->page2));
+	sp->page2 = calloc(1, sizeof(struct ses_page));
 	if (!sp->page2)
 		goto sp2;
-	sp->page10 = calloc(SES_ALLOC_BUFF, sizeof(*sp->page10));
-	if (!sp->page10)
-		goto sp10;
 	return sp;
-
- sp10:
-	free(sp->page2);
  sp2:
 	free(sp->page1);
  sp1:
@@ -161,13 +126,9 @@ static void ses_free(struct ses_pages *sp)
 {
 	if (!sp)
 		return;
-	if (sp->page1)
-		free(sp->page1);
-	if (sp->page2)
-		free(sp->page2);
-	if (sp->page10)
-		free(sp->page10);
-	memset(sp, 0, sizeof(*sp));
+	free(sp->page1);
+	free(sp->page2);
+	free(sp->page10);
 	free(sp);
 }
 
@@ -183,232 +144,103 @@ static void dump_p10(unsigned char *p)
 	}
 }
 
-/* Enclosure is already opened. */
-static int is_addr_in_encl(int fd, const char *addr, int *idx)
+static int enclosure_open(const struct enclosure_device *enclosure)
 {
-	/* Get page10 and read address. If it fits, return 1. */
-	struct ses_pages *sp = ses_init();
-	unsigned char *add_desc = NULL, *types = NULL;
-	unsigned char *ap = NULL, *addr_p = NULL;
-	int i, j, len = 0;
-	char addr_cmp[32];
-
-	if (!sp)
-		return 0;
-	/* Start processing */
-	/* Read configuration. */
-	if (get_page1(fd, sp->page1, &sp->page1_len))
-		goto err;
-
-	if (process_page1(sp))
-		goto err;
-
-	/* Get Enclosure Status - needed? */
-	if (get_page2(fd, sp->page2, &sp->page2_len))
-		goto err;
-
-	/* Additional Element Status */
-	if (get_page10(fd, sp->page10, &sp->page10_len))
-		goto err;
-
-	if (debug)
-		print_page10(sp);
-
-	/* Check Page10 for address. Extract index. */
-	ap = add_desc = sp->page10 + 8;
-	types = sp->page1_types;
-	for (i = 0; i < sp->page1_types_len; i++, types += 4) {
-		if (types[0] == SES_DEVICE_SLOT ||
-		    types[0] == SES_ARRAY_DEVICE_SLOT) {
-			for (j = 0; j < types[1]; j++, ap += len) {
-				if (debug)
-					dump_p10(ap);
-				/* Get Additional Element Status Descriptor */
-				/* length (x-1) */
-				len = ap[1] + 2;
-				if ((ap[0] & 0xf) != SCSI_PROTOCOL_SAS)
-					continue;	/* need SAS PROTO */
-				/* It is a SAS protocol, go on */
-				if ((ap[0] & 0x10))	/* Check EIP */
-					addr_p = ap + 8;
-				else
-					addr_p = ap + 4;
-				/* Process only PHY 0 descriptor. */
-				sprintf(addr_cmp,
-					"%02x%02x%02x%02x%02x%02x%02x%02x",
-					addr_p[12], addr_p[13], addr_p[14],
-					addr_p[15], addr_p[16], addr_p[17],
-					addr_p[18], addr_p[19]);
-
-				if (!strcmp(addr + 2, addr_cmp)) {
-					if (idx)
-						*idx = ap[3];
-					ses_free(sp);
-					return 1;
-				}
-			}
-		}
-	}
- err:
-	ses_free(sp);
-	return 0;
-}
-
-static char *_get_dev_sg(const char *path)
-{
-	char *ret = NULL;
-	DIR *d;
-	struct dirent *de;
-
-	/* /sys/class/enclosure/X/device/scsi_generic path is expected. */
-
-	d = opendir(path);
-	if (!d)
-		return NULL;
-	while ((de = readdir(d))) {
-		if ((strcmp(de->d_name, ".") == 0) ||
-		    (strcmp(de->d_name, "..")) == 0) {
-			continue;
-		}
-		break;
-		/* */
-	}
-	if (de) {
-		ret = calloc(strlen("/dev/") +
-			     strlen(de->d_name) + 1, sizeof(*ret));
-		if (ret) {
-			strcpy(ret, "/dev/");
-			strcat(ret, de->d_name);
-		}
-	}
-	closedir(d);
-	return ret;
-}
-
-/* SYSFS_ENCL/<enclosure>/SCSI_GEN - path to sgX directory for enclosure. */
-#define SYSFS_ENCL "/sys/class/enclosure"
-#define SCSI_GEN "device/scsi_generic"
-static int get_enclosure_fd(struct block_device *device, char *addr)
-{
-	DIR *d = NULL;
-	struct dirent *de = NULL;
-	char *p = NULL;
-	int len;
-	char *dev = NULL;
 	int fd = -1;
 
-	/* There may be device path already filled */
-	if (device->encl_index != -1 && device->encl_dev[0] != 0) {
-		fd = open(device->encl_dev, O_RDWR);
-		if (fd == -1) {
-			/* Enclosure device may change during extensive disks
-			 * hotplugging */
-			device->encl_index = -1;
-			memset(device->encl_dev, 0, sizeof(device->encl_dev));
+	if (enclosure->dev_path)
+		fd = open(enclosure->dev_path, O_RDWR);
 
-			if (!addr) {
-				/* If there is no SAS address then there is
-				 * no way to find enclosure device that this
-				 * drive 'was' in. */
-				return fd;
-			}
-		} else {
-			return fd;
-		}
-	}
-
-	if (!addr) {
-		/* when there is no enclosure device and device index
-		 * then address should not be NULL */
-		return -1;
-	}
-
-	d = opendir(SYSFS_ENCL);
-	if (!d)
-		return -1;
-	while ((de = readdir(d))) {
-		if ((strcmp(de->d_name, ".") == 0) ||
-		    (strcmp(de->d_name, "..")) == 0) {
-			continue;
-		}
-		/* */
-		len =
-		    strlen(SCSI_GEN) + strlen(SYSFS_ENCL) + strlen(de->d_name) +
-		    4;
-		p = calloc(len, sizeof(*p));
-		if (!p)
-			break;
-		strcpy(p, SYSFS_ENCL);
-		strcat(p, "/");
-		strcat(p, de->d_name);
-		strcat(p, "/");
-		strcat(p, SCSI_GEN);
-		dev = _get_dev_sg(p);
-		free(p);
-		if (!dev)
-			break;
-		fd = open(dev, O_RDWR);
-		if (fd == -1) {
-			free(dev);
-			break;
-		}
-		if (is_addr_in_encl(fd, addr, &device->encl_index)) {
-			strncpy(device->encl_dev, dev, PATH_MAX);
-			free(dev);
-			break;	/* HIT */
-		} else {
-			close(fd);
-			fd = -1;
-		}
-		free(dev);
-	}
-	closedir(d);
 	return fd;
 }
 
-static void put_enclosure_fd(int fd)
+static int enclosure_load_pages(struct enclosure_device *enclosure)
 {
-	close(fd);
-}
-
-static void ses_send_to_idx(int fd, int idx, enum ibpi_pattern ibpi)
-{
-	unsigned char msg[4] = { 0 };
+	int ret;
+	int fd;
 	struct ses_pages *sp;
 
+	if (enclosure->ses_pages)
+		return 0;
+
+	fd = enclosure_open(enclosure);
+	if (fd == -1)
+		return 1;
+
 	sp = ses_init();
-	if (!sp)
-		return;
+	if (!sp) {
+		ret = 1;
+		goto end;
+	}
 
-	/* Start processing */
 	/* Read configuration. */
-	if (get_page1(fd, sp->page1, &sp->page1_len))
-		goto err;
+	ret = get_ses_page(fd, sp->page1, ENCL_CFG_DIAG_STATUS);
+	if (ret)
+		goto end;
 
-	if (process_page1(sp))
-		goto err;
+	ret = process_page1(sp);
+	if (ret)
+		goto end;
 
 	/* Get Enclosure Status */
-	if (get_page2(fd, sp->page2, &sp->page2_len))
-		goto err;
+	ret = get_ses_page(fd, sp->page2, ENCL_CTRL_DIAG_STATUS);
+end:
+	close(fd);
+	if (ret)
+		ses_free(sp);
+	else
+		enclosure->ses_pages = sp;
+	return ret;
+}
 
-	if (ses_set_message(ibpi, msg))
-		goto err;	/* unknown message */
+static int enclosure_load_page10(struct enclosure_device *enclosure)
+{
+	int ret;
+	int fd;
+	struct ses_page *p;
 
-	send_diag_slot(sp, fd, msg, idx);
+	if (enclosure->ses_pages && enclosure->ses_pages->page10)
+		return 0;
 
- err:
-	ses_free(sp);
-	return;
+	ret = enclosure_load_pages(enclosure);
+	if (ret)
+		return ret;
+
+	fd = enclosure_open(enclosure);
+	if (fd == -1)
+		return 1;
+
+	p = calloc(1, sizeof(struct ses_page));
+	if (!p) {
+		ret = 1;
+		goto end;
+	}
+
+	/* Additional Element Status */
+	ret = get_ses_page(fd, p, ENCL_ADDITIONAL_EL_STATUS);
+end:
+	close(fd);
+	if (ret)
+		free(p);
+	else
+		enclosure->ses_pages->page10 = p;
+
+	return ret;
+}
+
+static void enclosure_free_pages(struct enclosure_device *enclosure)
+{
+	ses_free(enclosure->ses_pages);
+	enclosure->ses_pages = NULL;
 }
 
 static void print_page10(struct ses_pages *sp)
 {
-	unsigned char *ai = sp->page10 + 8;
+	unsigned char *ai = sp->page10->buf + 8;
 	int i = 0, len = 0, eip = 0;
 	unsigned char *sas = NULL;
 
-	while (ai < sp->page10 + sp->page10_len) {
+	while (ai < sp->page10->buf + sp->page10->len) {
 		printf("%s()[%d]: Inv: %d, EIP: %d, Proto: 0x%04x\n", __func__,
 		       i++, ((ai[0] & 0x80) >> 7), ((ai[0] & 0x10) >> 4),
 		       ai[0] & 0xf);
@@ -466,138 +298,184 @@ static void print_page10(struct ses_pages *sp)
 	return;
 }
 
-static void send_diag_slot(struct ses_pages *sp, int fd,
-			   const unsigned char *info, int idx)
-{
-	unsigned char *types = sp->page1_types;
-	unsigned char *desc = sp->page2 + 8;	/* Move do descriptors */
-	int i, j;
-	int slot = 0;
-
-	memset(desc, 0, sp->page2_len - 8);
-
-	for (i = 0; i < sp->page1_types_len; i++, types += 4) {
-		if (types[0] == 0x01 || types[0] == 0x17) {
-			desc += 4;	/* At first, skip overall header. */
-			for (j = 0; j < types[1]; j++, desc += 4) {
-				if (slot++ == idx) {
-					memcpy(desc, info, 4);
-					/* set select */
-					desc[0] |= 0x80;
-					/* keep PRDFAIL */
-					desc[0] |= 0x40;
-					/* clear reserved flags */
-					desc[0] &= 0xf0;
-					break;
-				}
-			}
-		}
-	}
-	if (sg_ll_send_diag(fd, 0, 1, 0, 0, 0, 0,
-			    sp->page2, sp->page2_len, 0, debug)) {
-		return;
-	}
-	return;
-}
-
-int ses_set_message(enum ibpi_pattern ibpi, unsigned char *u)
+static enum ibpi_pattern ibpi_to_ses(enum ibpi_pattern ibpi)
 {
 	switch (ibpi) {
 	case IBPI_PATTERN_UNKNOWN:
 	case IBPI_PATTERN_ONESHOT_NORMAL:
 	case IBPI_PATTERN_NORMAL:
-		_clr_msg(u);
-		_set_ok(u);
-		break;
+		return SES_REQ_OK;
 	case IBPI_PATTERN_FAILED_ARRAY:
-		_set_ifa(u);
-		break;
+		return SES_REQ_IFA;
 	case IBPI_PATTERN_DEGRADED:
-		_set_ica(u);
-		break;
+		return SES_REQ_ICA;
 	case IBPI_PATTERN_REBUILD:
-	case IBPI_PATTERN_REBUILD_P:
-		_set_rebuild(u);
-		break;
+		return SES_REQ_REBUILD;
 	case IBPI_PATTERN_FAILED_DRIVE:
-		_set_fault(u);
-		break;
-	case IBPI_PATTERN_LOCATE_OFF:
-		_clr_ident(u);
-		break;
+		return SES_REQ_FAULT;
 	case IBPI_PATTERN_LOCATE:
-		_set_ident(u);
-		break;
+		return SES_REQ_IDENT;
 	case IBPI_PATTERN_HOTSPARE:
-		_set_hspare(u);
-		break;
+		return SES_REQ_HOSTSPARE;
 	case IBPI_PATTERN_PFA:
-		_set_rsvd_dev(u);
-		break;
-		/* SES MESSAGES */
+		return SES_REQ_PRDFAIL;
+	default:
+		return ibpi;
+	}
+}
+
+static int ses_set_message(enum ibpi_pattern ibpi, struct ses_slot_ctrl_elem *el)
+{
+	struct ses_slot_ctrl_elem msg = { 0 };
+
+	if (ibpi == IBPI_PATTERN_LOCATE_OFF) {
+		/*
+		 * For locate_off we don't set a new state, just clear the
+		 * IDENT bit and the bits that are reserved or have different
+		 * meanings in Status and Control pages (RQST ACTIVE and
+		 * RQST MISSING).
+		 */
+		_clr_ident(el->b);
+		el->b2 &= 0x4e;
+		el->b3 &= 0x3c;
+		return 0;
+	}
+
+	switch (ibpi_to_ses(ibpi)) {
 	case SES_REQ_ABORT:
-		_set_abrt(u);
+		_set_abrt(msg.b);
 		break;
 	case SES_REQ_REBUILD:
-		_set_rebuild(u);
+		_set_rebuild(msg.b);
 		break;
 	case SES_REQ_IFA:
-		_set_ifa(u);
+		_set_ifa(msg.b);
 		break;
 	case SES_REQ_ICA:
-		_set_ica(u);
+		_set_ica(msg.b);
 		break;
 	case SES_REQ_CONS_CHECK:
-		_set_cons_check(u);
+		_set_cons_check(msg.b);
 		break;
 	case SES_REQ_HOSTSPARE:
-		_set_hspare(u);
+		_set_hspare(msg.b);
 		break;
 	case SES_REQ_RSVD_DEV:
-		_set_rsvd_dev(u);
+		_set_rsvd_dev(msg.b);
 		break;
 	case SES_REQ_OK:
-		_set_ok(u);
+		_set_ok(msg.b);
 		break;
 	case SES_REQ_IDENT:
-		_set_ident(u);
+		_set_ident(msg.b);
 		break;
 	case SES_REQ_RM:
-		_set_rm(u);
+		_set_rm(msg.b);
 		break;
 	case SES_REQ_INS:
-		_set_ins(u);
+		_set_ins(msg.b);
 		break;
 	case SES_REQ_MISSING:
-		_set_miss(u);
+		_set_miss(msg.b);
 		break;
 	case SES_REQ_DNR:
-		_set_dnr(u);
+		_set_dnr(msg.b);
 		break;
 	case SES_REQ_ACTIVE:
-		_set_actv(u);
+		_set_actv(msg.b);
 		break;
 	case SES_REQ_EN_BB:
-		_set_enbb(u);
+		_set_enbb(msg.b);
 		break;
 	case SES_REQ_EN_BA:
-		_set_enba(u);
+		_set_enba(msg.b);
 		break;
 	case SES_REQ_DEV_OFF:
-		_set_off(u);
+		_set_off(msg.b);
 		break;
 	case SES_REQ_FAULT:
-		_set_fault(u);
+		_set_fault(msg.b);
+		break;
+	case SES_REQ_PRDFAIL:
+		_set_prdfail(msg.b);
 		break;
 	default:
 		return 1;
 	}
+
+	*el = msg;
+
 	return 0;
 }
 
-/* SAS_ADDR_PATH - path in sysfs where sas_address for disk can be found */
-#define SAS_ADDR_PATH "/sys/class/sas_end_device/%s/" \
-						"device/sas_device/%s/sas_address"
+static int ses_write_msg(enum ibpi_pattern ibpi, struct block_device *device)
+{
+	struct ses_pages *sp = device->enclosure->ses_pages;
+	int idx = device->encl_index;
+	/* Move do descriptors */
+	struct ses_slot_ctrl_elem *descriptors = (void *)(sp->page2->buf + 8);
+	int i;
+	struct ses_slot_ctrl_elem *desc_element = NULL;
+	element_type element_type = SES_UNSPECIFIED;
+
+	for (i = 0; i < sp->page1_types_len; i++) {
+		struct type_descriptor_header *t = &sp->page1_types[i];
+
+		descriptors++; /* At first, skip overall header. */
+
+		if (t->element_type == SES_DEVICE_SLOT ||
+		    t->element_type == SES_ARRAY_DEVICE_SLOT) {
+			if (element_type < t->element_type &&
+			    t->num_of_elements > idx) {
+				element_type = t->element_type;
+				desc_element = &descriptors[idx];
+			}
+		} else {
+			/*
+			 * Device Slot and Array Device Slot elements are
+			 * always first on the type descriptor header list
+			 */
+			break;
+		}
+
+		descriptors += t->num_of_elements;
+	}
+
+	if (desc_element) {
+		int ret = ses_set_message(ibpi, desc_element);
+		if (ret)
+			return ret;
+		/* keep PRDFAIL, clear rest */
+		desc_element->common_control &= 0x40;
+		/* set select */
+		desc_element->common_control |= 0x80;
+
+		/* second byte is valid only for Array Device Slot */
+		if (element_type != SES_ARRAY_DEVICE_SLOT)
+			desc_element->array_slot_control = 0;
+
+		return 0;
+	}
+
+	return 1;
+}
+
+static int ses_send_diag(struct enclosure_device *enclosure)
+{
+	int ret;
+	int fd;
+
+	fd = enclosure_open(enclosure);
+	if (fd == -1)
+		return 1;
+
+	ret = sg_ll_send_diag(fd, 0, 1, 0, 0, 0, 0,
+			      enclosure->ses_pages->page2->buf,
+			      enclosure->ses_pages->page2->len,
+			      0, debug);
+	close(fd);
+	return ret;
+}
 
 static char *get_drive_end_dev(const char *path)
 {
@@ -614,53 +492,112 @@ static char *get_drive_end_dev(const char *path)
 		return NULL;
 
 	strncpy(p, c, s - c);
-	p[s - c] = 0;
 	return p;
 }
 
-static char *get_drive_sas_addr(const char *path)
+static uint64_t get_drive_sas_addr(const char *path)
 {
-#define ADDR_LEN 64
-	int len = strlen(path);
-	char *buff, *end_dev, addr[ADDR_LEN] = { 0 };
-	int fd;
+	uint64_t ret = 0;
+	size_t size = strlen(path) * 2;
+	char *buff, *end_dev;
 
 	/* Make big buffer. */
-	buff = calloc(len * 2, sizeof(*buff));
-
+	buff = malloc(size + 1);
 	if (!buff)
-		return NULL;
+		return ret;
 
 	end_dev = get_drive_end_dev(path);
 	if (!end_dev) {
 		free(buff);
-		return NULL;
+		return ret;
 	}
 
-	(void)sprintf(buff, SAS_ADDR_PATH, end_dev, end_dev);
+	snprintf(buff, size, "/sys/class/sas_end_device/%s/device/sas_device/%s",
+		 end_dev, end_dev);
+
+	ret = get_uint64(buff, ret, "sas_address");
 
-	fd = open(buff, O_RDONLY);
-	if (fd == -1) {
-		free(end_dev);
-		free(buff);
-		return NULL;
-	}
-	if (read(fd, addr, ADDR_LEN) == ADDR_LEN) {
-		/* The value should be 19. */
-		free(end_dev);
-		free(buff);
-		close(fd);
-		return NULL;
-	}
-	close(fd);
-	len = strnlen(addr, ADDR_LEN);
-	if (len && addr[len - 1] == '\n')
-		addr[len - 1] = 0;
-	else /* make sure that addr is null-terminated */
-		addr[len < ADDR_LEN ? len : ADDR_LEN - 1] = 0;
 	free(end_dev);
 	free(buff);
-	return strdup(addr);
+
+	return ret;
+}
+
+static int get_encl_slot(struct block_device *device)
+{
+	struct ses_pages *sp;
+	unsigned char *add_desc = NULL;
+	unsigned char *ap = NULL, *addr_p = NULL;
+	int i, j, len = 0;
+	uint64_t addr, addr_cmp;
+	int idx;
+
+	/* try to get slot from sysfs */
+	idx = get_int(device->cntrl_path, -1, "slot");
+	if (idx != -1)
+		return idx;
+
+	/*
+	 * Older kernels may not have the "slot" sysfs attribute,
+	 * fallback to Page10 method.
+	 */
+	if (enclosure_load_page10(device->enclosure))
+		return -1;
+	sp = device->enclosure->ses_pages;
+
+	addr = get_drive_sas_addr(device->sysfs_path);
+	if (!addr)
+		return -1;
+
+	if (debug)
+		print_page10(sp);
+
+	/* Check Page10 for address. Extract index. */
+	ap = add_desc = sp->page10->buf + 8;
+	for (i = 0; i < sp->page1_types_len; i++) {
+		struct type_descriptor_header *t = &sp->page1_types[i];
+
+		if (t->element_type == SES_DEVICE_SLOT ||
+		    t->element_type == SES_ARRAY_DEVICE_SLOT) {
+			for (j = 0; j < t->num_of_elements; j++, ap += len) {
+				if (debug)
+					dump_p10(ap);
+				/* Get Additional Element Status Descriptor */
+				/* length (x-1) */
+				len = ap[1] + 2;
+				if ((ap[0] & 0xf) != SCSI_PROTOCOL_SAS)
+					continue;	/* need SAS PROTO */
+				/* It is a SAS protocol, go on */
+				if ((ap[0] & 0x10))	/* Check EIP */
+					addr_p = ap + 8;
+				else
+					addr_p = ap + 4;
+				/* Process only PHY 0 descriptor. */
+
+				/* Convert be64 to le64 */
+				addr_cmp = ((uint64_t)addr_p[12] << 8*7) |
+					   ((uint64_t)addr_p[13] << 8*6) |
+					   ((uint64_t)addr_p[14] << 8*5) |
+					   ((uint64_t)addr_p[15] << 8*4) |
+					   ((uint64_t)addr_p[16] << 8*3) |
+					   ((uint64_t)addr_p[17] << 8*2) |
+					   ((uint64_t)addr_p[18] << 8*1) |
+					   ((uint64_t)addr_p[19]);
+
+				if (addr == addr_cmp) {
+					idx = ap[0] & 0x10 ? ap[3] : j;
+					return idx;
+				}
+			}
+		} else {
+			/*
+			 * Device Slot and Array Device Slot elements are
+			 * always first on the type descriptor header list
+			 */
+			break;
+		}
+	}
+	return -1;
 }
 
 /**
@@ -682,49 +619,47 @@ static int _slot_match(const char *slot_path, const char *device_path)
  */
 static char *_slot_find(const char *enclo_path, const char *device_path)
 {
-	void *dir;
+	struct list dir;
 	char *temp, *result = NULL;
 
-	dir = scan_dir(enclo_path);
-	if (dir) {
-		temp = list_first_that(dir, _slot_match, device_path);
-		if (temp)
-			result = strdup(temp);
-		list_fini(dir);
+	if (scan_dir(enclo_path, &dir) == 0) {
+		list_for_each(&dir, temp) {
+			if (_slot_match(temp, device_path)) {
+				result = strdup(temp);
+				break;
+			}
+		}
+		list_erase(&dir);
 	}
 	return result;
 }
 
 int scsi_get_enclosure(struct block_device *device)
 {
-	char *addr;
-	int fd = -1;
+	struct enclosure_device *encl;
 
 	if (!device || !device->sysfs_path)
 		return 0;
 
-	memset(device->encl_dev, 0, sizeof(device->encl_dev));
-	addr = get_drive_sas_addr(device->sysfs_path);
-	if (addr == NULL)
-		return 0;
+	list_for_each(sysfs_get_enclosure_devices(), encl) {
+		if (_slot_match(encl->sysfs_path, device->cntrl_path)) {
+			device->enclosure = encl;
+			device->encl_index = get_encl_slot(device);
+			break;
+		}
+	}
 
-	fd = get_enclosure_fd(device, addr);
-	if (fd != -1)
-		put_enclosure_fd(fd);
-	if (addr)
-		free(addr);
-
-	return (fd == -1) ? 0 : 1;
+	return (device->enclosure != NULL && device->encl_index != -1);
 }
 
 /**
  */
 int scsi_ses_write(struct block_device *device, enum ibpi_pattern ibpi)
 {
-	int fd = -1;
-	char *addr = NULL;
+	int ret;
 
-	if (!device || !device->sysfs_path)
+	if (!device || !device->sysfs_path || !device->enclosure ||
+	    device->encl_index == -1)
 		__set_errno_and_return(EINVAL);
 
 	/* write only if state has changed */
@@ -734,42 +669,45 @@ int scsi_ses_write(struct block_device *device, enum ibpi_pattern ibpi)
 	if ((ibpi < IBPI_PATTERN_NORMAL) || (ibpi > SES_REQ_FAULT))
 		__set_errno_and_return(ERANGE);
 
-	/* Failed drive is special. Path in sysfs may be not available.
-	 * In other case re-read address.
-	 * */
-	if (ibpi != IBPI_PATTERN_FAILED_DRIVE) {
-		addr = get_drive_sas_addr(device->sysfs_path);
-		if (addr == NULL) {
-			/* Device maybe gone during scan. */
-			log_warning
-			    ("Detected inconsistency. " \
-			     "Marking device '%s' as failed.",
-			     strstr(device->sysfs_path, "host"));
-			ibpi = IBPI_PATTERN_FAILED_DRIVE;
-			device->ibpi = IBPI_PATTERN_FAILED_DRIVE;
-
-			/* FIXME: at worst case we may lose all reference
-			 * to this drive and should remove it from list.
-			 * No API for do that now. */
-		}
-	}
-	fd = get_enclosure_fd(device, addr);
-	if (fd != -1) {
-		ses_send_to_idx(fd, device->encl_index, ibpi);
-		put_enclosure_fd(fd);
-	} else {
+	ret = enclosure_load_pages(device->enclosure);
+	if (ret) {
 		log_warning
 		    ("Unable to send %s message to %s. Device is missing?",
 		     ibpi_str[ibpi], strstr(device->sysfs_path, "host"));
+		return ret;
 	}
-	if (addr)
-		free(addr);
-	return 0;
+
+	return ses_write_msg(ibpi, device);
+}
+
+int scsi_ses_flush(struct block_device *device)
+{
+	int ret;
+
+	if (!device || !device->enclosure)
+		return 1;
+
+	if (!device->enclosure->ses_pages)
+		return 0;
+
+	ret = ses_send_diag(device->enclosure);
+
+	enclosure_free_pages(device->enclosure);
+	return ret;
 }
 
 /**
+ * @brief Gets a path to slot of sas controller.
+ *
+ * This function returns a sysfs path to component of enclosure the device
+ * belongs to.
+ *
+ * @param[in]      path           Canonical sysfs path to block device.
+ *
+ * @return A sysfs path to controller device associated with the given
+ *         block device if successful, otherwise NULL pointer.
  */
-char *sas_get_slot_path(const char *path, const char *ctrl_path)
+static char *sas_get_slot_path(const char *path, const char *ctrl_path)
 {
 	char *host;
 	char host_path[PATH_MAX] = { 0 };
@@ -793,12 +731,10 @@ static char *_get_enc_slot_path(const char *path)
 	struct enclosure_device *device;
 	char *result = NULL;
 
-	device = sysfs_get_enclosure_devices();
-	while (device) {
+	list_for_each(sysfs_get_enclosure_devices(), device) {
 		result = _slot_find(device->sysfs_path, path);
 		if (result != NULL)
 			break;
-		device = list_next(device);
 	}
 	return result;
 }
